import { KEY_PREFIX, REHYDRATE } from './constants';
import stringify from 'json-stringify-safe';

export default function createPersistoid(config) {
  // defaults
  var blacklist = config.blacklist || null;
  var whitelist = config.whitelist || null;
  var transforms = config.transforms || [];
  var throttle = config.throttle || 0;
  var storageKey = '' + (config.keyPrefix !== undefined ? config.keyPrefix : KEY_PREFIX) + config.key;

  // storage with keys -> getAllKeys for localForage support
  var storage = config.storage;
  if (storage.keys && !storage.getAllKeys) {
    storage.getAllKeys = storage.keys;
  }

  // initialize stateful values
  var lastState = {};
  var keysToProcess = [];
  var timeIterator = null;

  var update = function update(state) {
    Object.keys(state).forEach(function (key) {
      var subState = state[key];
      if (!passWhitelistBlacklist(key)) return; // is keyspace ignored? noop
      if (lastState[key] === state[key]) return; // value unchanged? noop
      if (keysToProcess.indexOf(key) !== -1) return; // is key already queued? noop
      keysToProcess.push(key); // add key to queue
    });

    // time iterator (read: throttle)
    if (timeIterator === null) {
      timeIterator = setInterval(function () {
        if (keysToProcess.length === 0) {
          if (timeIterator) clearInterval(timeIterator);
          timeIterator = null;
          return;
        }

        var key = keysToProcess.shift();
        var endState = transforms.reduce(function (subState, transformer) {
          return transformer.in(subState, key);
        }, lastState[key]);
        if (typeof endState !== 'undefined') stagedWrite(key, endState);
      }, throttle);
    }

    lastState = state;
  };

  var stagedState = {};
  function stagedWrite(key, endState) {
    stagedState[key] = serializer(endState);
    if (keysToProcess.length === 0) {
      storage.setItem(storageKey, serializer(stagedState), onWriteFail);
    }
  }

  function passWhitelistBlacklist(key) {
    if (whitelist && whitelist.indexOf(key) === -1) return false;
    if (blacklist && blacklist.indexOf(key) !== -1) return false;
    return true;
  }

  function onWriteFail() {
    return function setError(err) {
      // @TODO add fail handlers (typically storage full)
      if (err && process.env.NODE_ENV !== 'production') {
        console.error('Error storing data', err);
      }
    };
  }

  // return `persistoid`
  return {
    update: update
  };
}

function serializer(data) {
  return stringify(data, null, null, function (k, v) {
    if (process.env.NODE_ENV !== 'production') {
      throw new Error('\n        redux-persist: cannot process cyclical state.\n        Consider changing your state structure to have no cycles.\n        Alternatively blacklist the corresponding reducer key.\n        Cycle encounted at key "' + k + '" with value "' + v + '".\n      ');
    }
    return null;
  });
}

function deserializer(serial) {
  return JSON.parse(serial);
}